/*				EVENTS.C
*
*	This file contains the tracing facilities using the event data type.
*	It also contains the output routines for printing both proofs and
*	diagnostics, and the corresponding initialisations.
*
*	Note that the verbosity is defined:
*	0	Print no statistics at all
*	1	Print just the profile and Hall of Fame at the end
*	2	Print the profile for each sequent as well
*	3	Print the log of individual events and all profiles
*
*	The profile is graphed iff the -g switch is on.
*/

#include "minlog.h"


static
int	last_formula,	/* The formula most recently printed		*/
	last_line;	/* The current line number			*/


void put_bar( FILE *outdat, int offset, float height, int scale );
void put_axes( FILE *outdat, int maximum, int scale );



/*
*	The following initialisation function is too boring to be 
*	worth any real commentary.
*/
	
void events_init()
{
	event e;

	for ( e = try; e < max_event; e++ ) {
		profile[e] = 0;
		if ( !seq_number ) prof_sum[e] = 0;
	}
	event_string[try] = "Entry:";		event_string[member] = "Axiom:";
	event_string[bot] = "FE:";		event_string[ifI] = "->I:";
	event_string[andE] = "&E:";		event_string[MPP] = "MPP:";		
	event_string[TN] = "~~~:";		event_string[redundant] = "Redund:";	
	event_string[absurd] = "absurd";	event_string[share] = "Irrelev:";	
	event_string[andI] = "&I:";		event_string[orE] = "VE:";		
	event_string[gotT] = "Good:";		event_string[gotF] = "Bad:";		
	event_string[inbr] = "In_Br:";		event_string[B2F] = "-B2:";		
	event_string[orIT] = "+VI:";		event_string[orIF] = "-VI:";
	event_string[CP1F] = "-CP1:";		event_string[CP2F] = "-CP2:";
	event_string[ifE1] = "+aro_1:";		event_string[ifE2] = "+aro_2:";
	event_string[nogood] = "No_good:";

	rule_name[norule] = "<NULL>";		rule_name[inspect] = "inspection";
	rule_name[iI] = "arrow_I";		rule_name[aE] = "&E";
	rule_name[TNe] = "Double_Negation";	rule_name[vI] = "vI";
	rule_name[aI] = "&I";			rule_name[vE] = "vE";
	rule_name[iE] = "arrow_E";		rule_name[has] = "already_proved";
	rule_name[ASS] = "ASSUMPTION";		rule_name[REIT] = "REIT";
	rule_name[BotE] = "Bot_E";
	rule_name[n_rules] = "<ERROR>";

	heap_ptr = 0;
	event_counter = 0;
}






/*
*	"Report" statements are included here and there in the prover
*	in order to generate a profile of its behaviour.
*	Prof_sum is the cumulative version of profile, the latter 
*	being re-set to zero during each initialisation (see above).
*/

void Report( event e, int nn )
{
	profile[e]++;
	prof_sum[e]++;
	event_counter++;
	if ( verbosity == 3 ) {
		printf("%3d %-8s  ", event_counter, event_string[e]);
		if ( e != try ) printf("%d\n", nn);
	}
}







/*
*	At the end of each proof the profile is printed out using
*	this function with parameter "profile".  At the end of the
*	execution the cumulative profile is printed out using this
*	function with the parameter "prof_sum".
*
*	See the definitions of verbosity above.
*/

void dump_events( event p[max_event] )
{
	event e;
	int i, events_left;
	float perca, percb, percc;
	FILE *outdat, *fopen();

	if ( verbosity ) {
	    printf("\nProfile print:\n\n");
	    printf(" Entry:%10d\n\n", (events_left = p[try]));
	    for ( e = 1; e < max_event; e++ )
		printf( " %-9s%8d\n", event_string[e], p[e] );
	}

	if ( p == prof_sum ) {
	    if ( verbosity ) {
		printf("\n\n Valid:   %d\n Invalid: %d\n\n", 
				valid_count, invalid_count);
		printf("Worst cases:\n\n       VALID	      INVALID\n\n");
		for ( i = 0; i < HOF; i++ ) 
		if ( hall_of_fame[0][i] || hall_of_fame[2][i] )
		printf( "%6d %4.2f sec %10d %4.2f sec\n", 
			hall_of_fame[0][i],
			(hall_of_fame[1][i] * 1.0)/1000000,
			hall_of_fame[2][i],
			(hall_of_fame[3][i] * 1.0)/1000000 );
	    }

	    if ( graphs ) {

		if ( !(outdat = fopen("bargraph.tex","w")) ) {
		    printf("\n Warning: can't open the TeX output file\n");
		    return;
		}
		fprintf(outdat, "\\documentstyle[12pt]{article}\n");
		fprintf(outdat, "\\oddsidemargin=15mm\n");
		fprintf(outdat, "\\evensidemargin=15mm\n");
		fprintf(outdat, "\\textwidth=160mm\n");
		fprintf(outdat, "\\topmargin=0mm\n");
		fprintf(outdat, "\\textheight=300mm\n");
		fprintf(outdat, "\\begin{document}\n\n");
		fprintf(outdat, "\\thicklines\n");
		fprintf(outdat, "\\setlength{\\unitlength}{2mm}\n\n");
		fprintf(outdat, "\\pagestyle{empty}\n");

		fprintf(outdat, "\\newcommand{\\Bar}[3]{\\put(#1,0){\\line(0,1){#2}}%%\n");
		fprintf(outdat, "\\put(#3,0){\\line(0,1){#2}}%%\n");
		fprintf(outdat, "\\put(#1,#2){\\line(1,0){3}}}\n");

		fprintf(outdat, "$$\\begin{picture}(80,100)(-5,0)\n");
		for ( e = 1; e < max_event; e++ )
		    put_bar( outdat, e-1, (p[e]*100.0)/p[try], 2 );
		put_axes( outdat, max_event, 2 );

		events_left = p[try];
		fprintf(outdat, "$$\\begin{picture}(80,100)(-5,0)\n");
		for ( e = 1; e < max_event-1; e++ ) {
		    if ( events_left ) 
			put_bar( outdat, e-1, (p[e]*100.0)/events_left, 2 );
		    else put_bar( outdat, e-1, 0, 2 );
		    events_left -= p[e];
		    
		}
		put_axes( outdat, max_event-1, 2 );
	
		events_left = p[try];
		fprintf(outdat, "$$\\begin{picture}(80,100)(-5,0)\n");
		for ( e = 1; e < max_event; e++ ) {
		    events_left -= p[e];
		    put_bar( outdat, e-1, (events_left*100.0)/p[try], 1 );
		}
		put_axes( outdat, max_event, 1 );
		fprintf( outdat, "\\end{document}\n");
		fclose(outdat);
	    }
	}
}



void put_bar( FILE *outdat, int offset, float height, int scale )
{
	height *= scale;
	offset *= 3;
	fprintf(outdat, "  \\Bar{%d}{%3.2f}{%d}\n", offset, height, offset+3);
}


void put_axes( FILE *outdat, int maximum, int scale )
{
	fprintf(outdat, "  \\put(0,0){\\line(1,0){%d}} ", maximum*3);
	fprintf(outdat, "  \\put(0,0){\\line(0,1){100}}\n");
	fprintf(outdat, "  \\put(-2,0){\\makebox(0,0)[r]{0}}\n");
	fprintf(outdat, "  \\put(-2,50){\\makebox(0,0)[r]{%d}}\n", 50/scale);
	fprintf(outdat, "  \\put(-2,100){\\makebox(0,0)[r]{%d}}\n", 100/scale);

	fprintf(outdat, "\\put(0,-1){\\line(0,-1){3}}");
	fprintf(outdat, "\\put(9,-1){\\line(0,-1){3}}\n");
	fprintf(outdat, "\\put(33,-1){\\line(0,-1){3}}");
	fprintf(outdat, "\\put(42,-1){\\line(0,-1){3}}\n");
	fprintf(outdat, "\\put(63,-1){\\line(0,-1){3}}\n");
	fprintf(outdat, "\\put(4.5,-2.5){\\makebox(0,0){trivia}}\n");
	fprintf(outdat, "\\put(19.5,-2.5){\\makebox(0,0){invertible rules}}\n");
	fprintf(outdat, "\\put(36,-2.5){\\makebox(0,0){filters}}\n");
	fprintf(outdat, "\\put(51,-2.5){\\makebox(0,0){non-invertible rules}}\n");

	fprintf(outdat, "\\end{picture}$$\n\n\\newpage\n\n");
}



/*
*	The deparser converts the abstract "wff" representation of
*	formulas back into strings.  "Deparse" (upper case D) does
*	this for sets of formulas, being called for example in the
*	process of dumping a sequent.  The "first" formula in a list
*	is not preceeded by a comma.
*/

void Deparse( set Gamma )
{
	int i, first = true;

	WITHALL(i) {
		if ( first ) first = false;
		else printf(", ");
		deparse(i);
	}
}




/*
*	This deparses a single formula, putting no parenthses 
*	around the whole.
*/

void deparse( int formula )
{
	sub_deparse( true, formula );
}




/*
*	The guts of the deparser is this recursive part for 
*	deparsing a subformula into infix notation.
*	The boolean parameter is set if we are at the top
*	(outermost) level where surrounding parentheses are
*	not needed.
*/

void sub_deparse( boolean top, int formula )
{
	if ( atomic(formula) ) putchar(wff[formula].sym);
	else if ( wff[formula].sym == '>' && wff[formula].right == 1 ) {
	        putchar('~');
		sub_deparse( false, wff[formula].left );
	}
	else {
	        if ( !top ) putchar('(');
		sub_deparse( false, wff[formula].left );
		putchar(' ');
		if ( wff[formula].sym == '>' ) putchar('-');
		printf("%c ", wff[formula].sym);
		sub_deparse( false, wff[formula].right );
		if ( !top ) putchar(')');
	}
}





/*
*	The next group of functions are just like the above ones
*	except that they deparse that formula into Polish (prefix)
*	notation rather than infix.  This time the effect of the
*	"top" flag is to insert a space between left and right
*	subformulas.
*/

void PDeparse( set Gamma )   /*** Deparse into Polish ***/
{
	int i, first = true;

	WITHALL(i) {
	        if ( first ) first = false;
		else printf(", ");
		pdeparse(i);
	}
}



void pdeparse( int formula )
{
	psub_deparse( formula, true );
}



void psub_deparse( int formula, boolean top )
{
	char c;

	c = wff[formula].sym;

	if ( is_atom( wff[formula].sym )) putchar(wff[formula].sym);
	else if ( wff[formula].sym == '>' && wff[formula].right == 1 ) {
	        if ( wff[wff[formula].left].sym == '>' &&
		wff[wff[formula].left].right == 1 &&
		wff[wff[wff[formula].left].left].sym == '&' ) {
		        putchar('*');
			psub_deparse( wff[wff[wff[formula].left].left].left, false );
			psub_deparse( wff[wff[wff[formula].left].left].right, false );
		}
		else {
		        putchar('~');
			psub_deparse( wff[formula].left, top );
		}
	}
	else {
	        putchar(wff[formula].sym);
		psub_deparse( wff[formula].left, false );
		if ( top ) putchar(' ');
		psub_deparse( wff[formula].right, false );
	}
}







/*
*	Now the routine to dump a Fitch-style natural deduction version
*	of the proof.  Credit where it is due: this method was pointed 
*	out to me by Seppo Keronen following a suggestion by Neil Tennant.
*
*	The first few lines report the premises of the sequent to be 
*	proved.  Subsequent ones may be indented to indicate sub-proof
*	status.  The rules &I and vE which require two sub-proofs help the
*	eye by inserting a blank line between the two parts.
*/

void proof_out( proof P, set Gamma )
{
	int i;

	if ( P ) {
		valid_count++;
		good_sequent = true;
		if ( printem) {
			WITHALL(i) print_node( i, ASS, 0 );
			print_proof( P, 0 );
			printf("\n\n\n");
		}
		else printf("\n Sequent %d is valid\n",seq_number);
	}
	else {
		invalid_count++;
		good_sequent = false;
		printf("\n Sequent %d is invalid\n",seq_number);
	}

	if ( verbosity > 1 ) dump_events( profile );
}




/*
*	The recursive part of printing a proof.  The parameter P is a
*	pointer to [the root of] a sub-proof to be printed out, while
*	the parameter indent records how many levels of sub-proof are
*	above this one.  The whole business is very case-ridden but 
*	not particularly complex.
*/

void print_proof( proof P, int indent )
{
	int i, j;

	if ( !P ) return;
	switch( P->rl ) {
		case norule:
		case n_rules: Abort("Bad rule citation");
		case has:
/*		case inspect: */
		case BotE:	print_node( P->form, P->rl, indent );
				break;
		case TNe:	print_node( P->form, TNe, indent );
				print_proof( P->lsprf, indent );
				break;
		case iI:	print_node( wff[P->form].left, ASS, indent+1 );
				print_proof( P->lsprf, indent+1 );
				print_node( P->form, iI, indent );
				break;
		case aE:	print_node( wff[P->form].left, aE, indent );
				print_node( wff[P->form].right, aE, indent );
				print_proof( P->lsprf, indent );
				break;
		case MP:	print_node( wff[P->form].right, MP, indent );
				print_proof( P->lsprf, indent );
				break;
		case or_if:	i = Loc( '>', wff[wff[P->form].left].left,
						wff[P->form].right );
				j = Loc( '>', wff[wff[P->form].left].right,
						wff[P->form].right );
				print_node( i, or_if, indent );
				print_node( j, or_if, indent );
				print_proof( P->lsprf, indent );
				break;
		case aI:	print_proof( P->lsprf, indent+1 );
				blank_line( indent );
				print_proof( P->rsprf, indent+1 );
				print_node( P->form, aI, indent );
				break;
		case vE:	print_node( P->form, REIT, indent );
				print_node( wff[P->form].left, ASS, indent+1 );
				print_proof( P->lsprf, indent+1 );
				blank_line( indent );
				print_node( wff[P->form].right, ASS, indent+1 );
				print_proof( P->rsprf, indent+1 );
				print_node( P->conc, vE, indent );
				break;
		case vI:	print_proof( P->lsprf, indent );
				print_node( P->form, vI, indent);
				break;
		case iE:	print_proof( P->lsprf, indent );
				print_node( wff[P->form].right, iE, indent );
				print_proof( P->rsprf, indent );
	}
}




/*
*	This subroutine just positions and prints a formula and the name
*	of a rule.  It can be set up to print in infix or prefix notaiton.
*/

void print_node( int fmla, rule rul, int indent )
{
	int i;

	if ( fmla == last_formula && rul == REIT ) return;
	for ( i = 0; i < indent; i++ ) printf(" | ");
	deparse( fmla );
	printf("   %s\n", rule_name[rul]);
	last_formula = fmla;
}




/*
*	This is trivial.  It is used in cases &I and vE above.
*/

void blank_line( int indent )
{
	int i;

	for ( i = 0; i < indent; i++ ) printf(" | ");
	puts("");
}
