/*			MINLOG.C: PREPROCESSOR
*
*	This file contains the pre-processing functions for the theorem
*	prover minlog.c.  After initialisation (called from main) the 
*	sequent to be proved is read from stdin (get_sequent is also 
*	called from main).  After parsing and re-parsing to simplify
*	certain compounds and to paraphrase away negation, the set of 
*	subformulas is sorted into an order of weight and the subf 
*	structure set up for rapid testing as in MaGIC.
*/


#include "minlog.h"



void prog_initial( int argc, char *argv[] )
{
	int option, i, j;
	extern char *optarg;
	extern int optind;

	printem = false;
	intuitionist = false;
	graphs = false;
	verbosity = 0;
	while ((option = getopt(argc, argv, "gipv:")) != -1)
	switch( option ) {
		case 'g': graphs = true; break;
		case 'i': intuitionist = true; break;
		case 'p': printem = true; break;
		case 'v': if ( !isdigit(*optarg) ) {
				fprintf(stderr,
                                "usage: minlog [-g] [-i] [-p] [-v number]\n");
				exit(1);
			}
			sscanf(optarg, "%d", &verbosity);
			if ( verbosity < 0 ) verbosity = 0;
			if ( verbosity > 3 ) verbosity = 3;
	}

	seq_number = 0;
	valid_count = invalid_count = 0;
	for ( i = 0; i < HOF; i++ )
	for ( j = 0; j < 4; j++ ) hall_of_fame[1][i] = hall_of_fame[3][i] = 0;
}



void initialise()
{
	int i, j, k;
	unsigned tim;

	zero = 0;

	events_init();
	for ( i = 0; i < MAXSUBS; i++ ) {
		wff[i].sym = '\0';
		wff[i].left = wff[i].right = 0;
		subf[i].lv = subf[i].rv = &zero;
		subf[i].val = 0;
	}
	wff[1].sym = 'F';
	ibmax = 0;
	if ( valids ) {
		subtract( valids );
		free( valids );
		valids = 0;
	}
	if ( invalids ) {
		subtract( invalids );
		free( invalids );
		invalids = 0;
	}

	con[0][0] = 0;  con[0][1] = 0;
	con[1][0] = 0;  con[1][1] = 1;

	dis[0][0] = 0;  dis[0][1] = 1;
	dis[1][0] = 1;  dis[1][1] = 1;

	aro[0][0] = 1;  aro[0][1] = 1;
	aro[1][0] = 0;  aro[1][1] = 1;

	tim = get_time();  k = (good_sequent? 1: 3);
	for ( i = 0; i < HOF; i++ )
	if ( hall_of_fame[k][i] < tim ) break;
	if ( i < HOF ) {
		for ( j = HOF-1; j > i; j-- ) {
			hall_of_fame[k-1][j] = hall_of_fame[k-1][j-1];
			hall_of_fame[k][j] = hall_of_fame[k][j-1];
		}
		hall_of_fame[k-1][i] = seq_number;
		hall_of_fame[k][i] = tim;
	}
}




void subtract( TREE *record )
{
	if ( record->leftsub ) {
		subtract( record->leftsub );
		free( record->leftsub );
	}
	if ( record->rightsub ) {
		subtract( record->rightsub );
		free( record->rightsub );
	}
	empty( record->prem );
	record->conc = 0;
	free( record->prem );
	record->leftsub = record->rightsub = 0;
}




boolean got_sequent( set Gamma, int *Con )
{
	if ( !insequent( Gamma, Con )) return( false );
	set_up_wff( Gamma, Con );

	if ( printem ) {
		Deparse( Gamma ); printf(" |- "); 
		deparse( *Con ); printf("\n\n");
	}

	set_up_subf();
	set_up_v_used();

	if ( verbosity == 3 ) dump_wff(Gamma,*Con);
}




boolean insequent( set Gamma, int *conclusion )
{
	char seq[512], fml[512];
	int i, j, k;

	seq_number++;
	if ( printem ) {
		printf("\n SEQUENT %d:  ", seq_number);
		fflush(stdout);
	}
	fgets(seq,512,stdin);
	if ( !*seq ) {
		if ( printem ) puts("< null input >");
		return( false );
	}
	j = *conclusion = 0;
	empty( Gamma );
	for ( i = 0; seq[i] == ' ' || seq[i] == ':'; i++ ) ;
	while ( true ) {
		for (	k = 0;
			seq[i] && seq[i] != ',' && seq[i] != ':' && seq[i] != '\n';
			k++)
		fml[k] = seq[i++];
		fml[k] = '\0';
		if ( seq[i] == '\n' ) seq[i] = '\0';
		if ( !seq[i++] ) {
			*conclusion = parse( fml );
			return( *conclusion? true: false );
		}
		else {
			if ( !(k = parse( fml )))
				return( false );
			putin( k, Gamma );
		}
	}
}




void set_up_wff( set Gamma, int *conclusion )
{
	int i, j, k;
	boolean b;

	for (	max_atoms = 1;
		is_atom(wff[max_atoms].sym); max_atoms++ );
	for (	max_subs = max_atoms;
		wff[max_subs].sym; max_subs++ );

	for ( i = max_atoms; i < max_subs; i++ )
	if ( is_atom( wff[i].sym )) {
		exchange( max_atoms, i, Gamma, conclusion );
		max_atoms++;
	}
	for ( i = 1; i < max_subs; i++ ) {
		if ( wff[i].left > i ) {
			exchange( i, wff[i].left, Gamma, conclusion );
			i = 0;
		}
		else if ( wff[i].right > i ) {
			exchange( i, wff[i].right, Gamma, conclusion );
			i = 0;
		}
	}


	p_mask = 0;
	for ( i = max_atoms; i < 32; i++ ) p_mask |= (1 << i);

	if ( max_atoms < 7 ) {
		j = 1;
		b = false;
		for ( i = (intuitionist? 2: 1); i < max_atoms; i++ ) {
			fives[i] = 0;
			for ( k = 0; k < 32; k++ ) {
				if ( !(k % j) ) b = !b;
				if ( b ) fives[i] |= (1 << k);
			}
			j *= 2;
		}
		if ( intuitionist ) fives[1] = 0;
		for ( i = max_atoms; i < max_subs; i++ )
		switch( wff[i].sym ) {
			case '&':
			fives[i] = fives[wff[i].left] & fives[wff[i].right];
			break;
			case 'V':
			fives[i] = fives[wff[i].left] | fives[wff[i].right];
			break;
			case '>':
			fives[i] =
				~(fives[wff[i].left]) | fives[wff[i].right];
		}
	}
}



void exchange( int lower, int higher, set Gamma, int *conclusion )
{
	int templeft, tempright;
	char tempsym;
	int i;

	templeft = wff[lower].left;
	tempright = wff[lower].right;
	tempsym = wff[lower].sym;
	wff[lower].left = wff[higher].left;
	wff[lower].right = wff[higher].right;
	wff[lower].sym = wff[higher].sym;
	wff[higher].left = templeft;
	wff[higher].right = tempright;
	wff[higher].sym = tempsym;
	for ( i = max_atoms; i < max_subs; i++ ) {
		if ( wff[i].left == lower ) wff[i].left = MAXSUBS;
		if ( wff[i].right == lower ) wff[i].right = MAXSUBS;
		if ( wff[i].left == higher ) wff[i].left = lower;
		if ( wff[i].right == higher ) wff[i].right = lower;
		if ( wff[i].left == MAXSUBS ) wff[i].left = higher;
		if ( wff[i].right == MAXSUBS ) wff[i].right = higher;
	}
	if ( in( lower, Gamma ) && !in( higher, Gamma )) {
		putin( higher, Gamma );
		putout( lower, Gamma );
	}
	else if ( in( higher, Gamma ) && !in( lower, Gamma )) {
		putin( lower, Gamma );
		putout( higher, Gamma );
	}
	if ( lower == *conclusion ) *conclusion = higher;
	else if ( higher == *conclusion ) *conclusion = lower;
}



void set_up_subf()
{
	int i;

	for ( i = 1; i < max_subs; i++ ) {
		subf[i].mtx =	( wff[i].sym == '&'? *con:
				( wff[i].sym == 'V'? *dis:
				( wff[i].sym == '>'? *aro: 0 )));
		subf[i].lv = &(subf[wff[i].left].val);
		subf[i].rv = &(subf[wff[i].right].val);
	}
}




void set_up_v_used()
{
	int i;

	for ( i = 1; wff[i].sym; i++ )
	v_used[i] = 
		atomic(i)? 1 << i:
		( v_used[wff[i].left] | v_used[wff[i].right] );
}




void dump_wff(set Gamma, int conclusion)
{
	int i;

	for ( i = 1; wff[i].sym; i++ ) {
		printf(" %d:  %d %c %d  ", 
			i, wff[i].left, wff[i].sym, wff[i].right);
		if ( i == max_atoms ) printf("m");
		if ( i == max_subs ) printf("M");
		if ( in( i, Gamma )) printf("P");
		if ( i == conclusion ) printf("C");
		printf("   ");
		pdeparse(i);
		puts("");
	}
}





void Abort( char *s )
{
	printf("\n\n Panic!!!  ");
	puts(s);
	printf("\n\n");
	exit(1);
}





dump_fives()
{
	int i, j;

	for ( i = 1; i < max_subs; i++ ) {
		printf("%2d  ",i);
		for ( j = 0; j < 16; j++ )
		printf("%d", ((fives[i] & (1 << j)) != 0));
		printf("   %d %c %d\n", wff[i].left, wff[i].sym,
			wff[i].right);
	}
}
