
/*				PROVER.C
*
*	This file contains the "proved" function which does most
*	of the work for the theorem prover minlog.c.
*
*	The first parameter is a pointer to the set of premises of
*	a sequent for which a proof is sought.  The second is its
*	conclusion.  The third is a superset of Gamma, containing
*	also some of the formulas already derived from Gamma higher
*	up the current branch of the proof search tree.  The fourth
*	is for recording any member of Gamma2-Gamma which gets used 
*	by a trivial rule.
*
*	The sequent is first reduced by dealing with "&" on both
*	right and left and with "v" on the left and "->" on the 
*	right.  At this time it is also checked to see whether 
*	the conclusion is also a premise and whether any such rule
*	as Modus Ponens applies trivially.  Then the sequent is 
*	checked against truth tables.  If it passes this filter
*	the databases of sequents already decided are consulted to 
*	see whether the present one is among them.  Then attempts are 
*	made to prove it by the rule for "v" on the right.  If all 
*	else fails the rule for "->" on the left is tried and the 
*	sequent added to one of the databases.
*
*	The function uses the macro "RETURN" to return a pointer to
*	the generated proof node if there is one.  Otherwise it just
*	returns a null value.
*/


#include "minlog.h"
#include "prover.h"


proof proved( set Gamma, int Con, set Gamma2, set Extras )
{
  register int i;
  int pi, pj, j, k;
  SET	DELTA, 		/* Used for the premises of subproofs	*/
    DELTA2,		/* Represents &Gamma2 for subproofs	*/
    SIGMA,		/* Implications on the left immune from
			   the rule ->E for various reasons	*/
    TAU;		/* For feedback as Extras		*/
  set	Delta,
    Delta2,
    Sigma,
    Tau;		/* These are just pointers to the above	*/
  int premises[MAXSUBS];	/* Members of Gamma for quick reference	*/
  proof subproofs[2];	/* Store for proof pointers generated	*/
  int	entry_count;



  report( try );		/* Note: "report" just logs events	*/
  entry_count = event_counter;

  NOLEFT;
  NORIGHT;		/* Initially there are no subproofs	*/
  Tau = &TAU;		empty( Tau );

  if ( verbosity == 3 ) {
    j = 0;
    WITHALL(i) printf("%d%c ", i, (j++? ',': '\0'));
    printf(" |-  ");
    deparse( Con );
    puts("");
  }


  /*
   *	First check for membership and apply the trivial rules.
   */
  if ( in( Con, Gamma2 )) {
    report( member );
    if ( !in( Con, Gamma )) putin( Con, Tau );
    RETURN( inspect, Con )
      }

  if ( intuitionist && in( Bot, Gamma2 )) {
    report( bot );
    if ( !in( Bot, Gamma )) putin( Bot, Tau );
    RETURN( BotE, Con )
      }

  if ( atomic(Con) && isempty(Gamma) ) {
    report( absurd );
    return(0);
  }

  Delta = &DELTA; 	Delta2 = &DELTA2;	Sigma = &SIGMA;
  RESET_DELTA

    /*
     *	Implication on the right is always a trivial rule.
     */
    if ( wff[Con].sym == '>' ) {
      report( ifI );
      NEWPREM( wff[Con].left )
	if ( L_SUBPROOF( wff[Con].right ))
	  RETURN( iI, Con )
	    return( 0 );
    }
  
  /*
   *	So is conjunction on the left.
   */
  WITH(i) {
    if ( wff[i].sym == '&' ) {
      report( andE );
      putout( i, Delta );
      NEWPREM( wff[i].left );
      NEWPREM( wff[i].right );
      if ( L_SUBPROOF( Con ))
	RETURN( aE, i )
	  return( 0 );
    }

    /*
     *	Now see whether Modus Ponens applies for free
     */
    if ( wff[i].sym == '>' && in( wff[i].left, Gamma2 )) {
      report( MPP );
      if ( !in( wff[i].left, Gamma )) putin( wff[i].left, Tau );
      NOLEFT;
      if ( wff[i].right == Con ) NORIGHT;
      else {
	putout( i, Delta );
	NEWPREM( wff[i].right )
	  if ( !(R_SUBPROOF( Con ))) return( 0 );
      }
      RETURN( iE, i )
	}
    /*
     *	Finally in this section, simplify triple negations
     *	and remove excess disjunctions and implications.
     *	Note that at this point *Gamma may be changed by the
     *	removal of some superfluous premises.
     */
    if ( wff[i].sym == '>' ) {
      if ( wff[i].right == Con && DOUBLE_NEG(i) ) {
	report( TN );
	putout( i, Delta );
	NEWPREM( wff[wff[i].left].left );
	if ( L_SUBPROOF( Con ))
	  RETURN( TNe, wff[wff[i].left].left )
	    return( 0 );
      }
      if ( in( wff[i].right, Gamma )) putout( i, Delta );
    }
    
    if ( wff[i].sym == 'V' &&
	 ( in( wff[i].left, Gamma ) || in( wff[i].right, Gamma )))
      putout( i, Delta );
    
  }
  if ( unequal( Gamma, Delta )) {
    report(redundant);
    return( L_SUBPROOF( Con ));
  }
  
  
  /*
   *	Minimal logic has a variable-sharing property: if there is
   *	no common variable between premises and conclusion then the 
   *	conclusion is a theorem.  In addition, any irrelevant atomic 
   *	premise is redundant.  The effect of  set_used  is that 
   *	p_used and c_used are the sets of atoms occurring in the 
   *	premises and conclusion respectively, while  used  is the
   *	set occurring in the conclusion or in non-atomic premises 
   *	together (for masking purposes) with the molecular formulas.
   */
  if ( !intuitionist ) {
    set_used( Gamma, Con );
    low_intersect( Delta, used );
    if ( !(p_used & c_used) )
      empty( Delta );
    if ( unequal( Gamma, Delta )) {
      report(share);
      return( L_SUBPROOF( Con ));
    }
  }
  
  /*
   *	The rules for & on the right and V on the left are still
   *	invertible, though they split the proof search, and should
   *	be tried next.
   */
  if ( wff[Con].sym == '&' ) {
    report( andI );
    if ( !(L_SUBPROOF( wff[Con].left ))) return 0;
    if ( !(R_SUBPROOF( wff[Con].right ))) return 0;
    RETURN( aI, Con )
      }
  
  WITH(i)
    if ( wff[i].sym == 'V' ) {
      report( orE );
      putout( i, Delta );
      NEWPREM( wff[i].left )
	if ( !(L_SUBPROOF( Con ))) return 0;
      RESET_DELTA
	putout( i, Delta );
      NEWPREM( wff[i].right )
	if ( !(R_SUBPROOF( Con ))) return 0;
      RETURN( vE, i )
	}
  
  /*
   *	*	*	*	*	*	*	*	*	*
   *
   *			T H E  G R E A T  D I V I D E
   *
   *	At this point we have a sequent which is not trivially
   *	reducible.  That is, the only remaining rules are the non-
   *	invertible ones.  Hence start by checking the databases.
   */
  
  if ( got( valids, Gamma, Con )) {
    report( gotT );
    RETURN( has, Con )
      }
  
  if ( got( invalids, Gamma, Con )) {
    report( gotF );
    return 0;
	}
  
  /*
   *	Next apply the matrix filter.  Since we are now through the
   *	database check, from this point on any return is preceded by
   *	addition of the sequent to the "valids" or to the "invalids".
   *	As a preliminary, copy Gamma into the array of premises.
   */
  j = 0;
  WITHALL(i) premises[j++] = i;
  premises[j] = 0;
  
  if ( !B2_valid( premises, Con, Gamma )) {
    report( B2F );
    ADD_FALSE
      return 0;
  }
  
  /*
   *	Now we have a genuinely new potentially valid sequent.
   *	If conclusion is disjunctive try each disjunct, then go on
   *	to deal with the really regrettable cases.
   */
  if ( wff[Con].sym == 'V' ) {
    if ( L_SUBPROOF( wff[Con].left )) {
      report( orIT );
      ADD_TRUE
	RETURN( vI, Con )
	}
    if ( L_SUBPROOF( wff[Con].right )) {
      report( orIT );
      ADD_TRUE
	RETURN( vI, Con )
	}
    if ( isempty( Gamma )) {
      ADD_FALSE
	report( orIF );
      return 0;
    }
  }
  
  /*
   *	The function inbranch checks whether the current sequent is
   *	among those under test (as yet undecided) on the present
   *	branch of the proof search tree.  This is necessary to prevent
   *	looping from this point.  Note that although this check may
   *	cause a return it does not directly bring about a decision
   *	that the sequent is unprovable, so there is no "add".
   */
  
  if ( inbranch( Gamma, Con )) {
    report( inbr );
    return 0;
  }
  
  putinbranch( Gamma, Con );
  
  /*
   *	In most cases  ->E  will work without keeping the eliminated
   *	implication in the premise set.  This is always the case with
   *	double negations in particular.  Sigmaa is used to treat all
   *	implications with the same antecedent together, since 
   *	{ A -> B, A -> C }  amounts to the same as { A -> (B & C) }.
   *	It is possible to look first at implications which contain
   *	the desired conclusion as an ultimate consequent part, but 
   *	the simpler version of the rule is used here.
   */
  
  empty( Sigma );
  P_WITH(pi,i) if ( !in( i, Sigma )) {
    RESET_DELTA
      P_WITH(pj,j)
      if ( wff[j].left == wff[i].left ) putout( j, Delta );
    if ( L_SUBPROOF( wff[i].left )) {
      RESET_DELTA
	P_WITH(pj,j) if ( wff[j].left == wff[i].left ) {
	if ( i != j ) putin( j, Sigma );
	putout( j, Delta );
	NEWPREM( wff[j].right )
	  }
      if ( in(Con, Delta2) ) NORIGHT;
      else if ( !(R_SUBPROOF( Con ))) {
	report( CP1F );
	ADD_FALSE
	  return( 0 );
      }
      report( ifE1 );
      ADD_TRUE
	RETURN( iE, i )
	}
    if ( DOUBLE_NEG(i) ) putin( i, Sigma );
  }
  
  P_WITH(pi,i) if ( !in( i, Sigma )) {
    RESET_DELTA
      if ( L_SUBPROOF( wff[i].left )) {
	RESET_DELTA
	  P_WITH(pj,j) if ( wff[j].left == wff[i].left ) {
	  putout( j, Delta );
	  NEWPREM( wff[j].right )
	    }
	if ( in(Con, Delta2) ) NORIGHT;
	else if ( !(R_SUBPROOF( Con ))) {
	  report( CP2F );
	  ADD_FALSE
	    return 0;
	}
	report( ifE2 );
	ADD_TRUE
	  RETURN( iE, i )
	  }
  }
  
  report( nogood );
  ADD_FALSE
    return( 0 );
}
