/*				TESTER.C
*
*	This file contains the routines for testing sequents for
*	validity in the matrices B2 (2-valued Boolean truth tables)
*	and H3 (the 3-element Heyting lattice).  The test module 
*	itself is pretty much as in MaGIC.
*
*	There is also a function "set_used" to determines which 
*	atoms occur in compound premises and in the conclusion.
*
*	Both set_used and the matrix tests are called from the 
*	"proved" function of the theorem prover minlog.
*/



#include "minlog.h"




/*
*	The following function sets up the external quantities
*
*		used		subformulas of conclusion and of
*				any compound (molecular) premises
*
*		c_used		atoms occurring in the conclusion
*
*		p_used		atoms occurring in premises
*
*	This is for the purpose of imposing appropriate variable-
*	sharing properties on the sequent under test.  Any atomic
*	premise is redundant unless it also occurs in a compound,
*	and if no atom is shared by premises and conclusion then
*	the sequent can only be provable if its conclusion is a
*	theorem.
*/

void set_used( set Gamma, int Con )
{
  register int i;
  halfset pc_used;

  p_used = c_used = pc_used = used = 0;
  WITHALL(i)
    if atomic( i ) p_used |= (1 << i);
    else pc_used |= v_used[i];
  c_used = v_used[Con];
  used = pc_used | c_used | p_mask;
  p_used |= pc_used;
}






/*
 *	The truth table test is intended to be fast rather than pretty,
 *	though the algorithm overall is the crudest possible, making
 *	all assignments of values to the atoms and spreading the
 *	valuation in each case to the compounds.  Formulas with many
 *	atoms would make it preferable to substitute a more efficient
 *	test.  Fortunately, this module is very self-contained, so
 *	such a substitution would be quite easy.
 *
 *	The first part of the function is setting up the test 
 *	structure.  This is a matter of deciding which subformulas
 *	occur and which atoms should be re-valued in the CHANGE part.
 *	Note that it is assumed that the subformula relation is 
 *	embedded in the numerical order of wff and subf.
 */

boolean B2_valid( int premises[MAXSUBS], int conclusion, set Gamma )
{
  unsigned s;
  int i;
  
  if ( max_atoms > 6 )
    return( B2_valid_test( premises, conclusion, Gamma ));
  
  s = ~( fives[conclusion] );
  for ( i = 0; premises[i]; i++ )
    s &= fives[premises[i]];
  return( !s );
}


boolean B2_valid_test( int premises[MAXSUBS], int conclusion, set Gamma )
{
  register TWFF *tptr;	/* For controlling the test loop	*/
  register int i;
  int	pi, j,		/* Inconsequential integers		*/
    start = 0;	/* First molecular formula used		*/
  SET All_used;		/* Set of subformulas to be evaluated	*/
  set all_used;		/* Simply a pointer to that set		*/


  all_used = &All_used;
  copyset( Gamma, all_used );

  if ( !atomic( conclusion )) putin( conclusion, all_used );
  for ( i = max_subs-1; i >= max_atoms; i-- ) if ( in(i, all_used) ) {
    if ( !start ) {
      start = i;
      subf[i].nxt = 0;
    }
    else {
      subf[i].nxt = subf+start;
      start = i;
    }
    putin( wff[i].left, all_used);
    putin( wff[i].right, all_used);
  }
  /*
   *    If there is no start position, there is nothing to test
   */
  if (!start)
    return false;
  /*
   *	That's collected all the subformulas to be evaluated.  Now any
   *	atomic premise is to be regarded as a true constant, and the
   *	conclusion, if it is atomic, as a false one.
   */
  for ( i = 0; atomic(premises[i]); i++ ) {
    subf[premises[i]].val = 1;
    putout( premises[i], all_used );
  }
  if ( atomic( conclusion )) {
    subf[conclusion].val = 0;
    putout( conclusion, all_used );
  }
  /*
   *	Now we are ready for the actual test-and-change loop.
   *	First the values of "live" variables are initialised to 0.
   *	Then tests and changes alternate until either a refuting
   *	assignment turns up or there is no further change to be made.
   *
   *	What the second line of the "test" part means is
   *		Value = Mat[x][y];
   *	where Mat is the matrix (conjunction, disjunction or implication)
   *	appropriate to a given subformula and Value is the value being
   *	assigned to it.  The integers x and y are the values already
   *	assigned to its subformulas.
   */
  for ( i = 1; i < max_atoms; i++ )
    if ( in( i, all_used )) subf[i].val = 0;
  
 TEST: for ( tptr = subf+start; tptr; tptr = tptr->nxt )
    tptr->val = *(tptr->mtx + (*(tptr->lv) << 1) + *(tptr->rv));
  
  if ( subf[conclusion].val ) goto CHANGE;
  for ( i = 0; premises[i]; i++ )
    if ( !subf[premises[i]].val ) goto CHANGE;
  return false;
  
 CHANGE: for ( j = (intuitionist? 2: 1); j < max_atoms; j++ )
    if ( in( j, all_used )) {
      if ( !subf[j].val ) {
	subf[j].val = 1;
	goto TEST;
      }
      subf[j].val = 0;
    }
  return true;
}
